/*
 * $Id: connection.c,v 1.19 2003/03/07 02:34:06 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gift.h"

#include "network.h"
#include "event.h"

#include "connection.h"

/*****************************************************************************/

struct write_msg
{
	unsigned char *data;
	size_t         len;
};

/*****************************************************************************/

static TCPConn *tcp_new (int fd, in_addr_t host, in_port_t port,
                         int outgoing, void *udata)
{
	TCPConn *c;

	if (!(c = MALLOC (sizeof (TCPConn))))
		return NULL;

	if (!(c->buf = fdbuf_new (fd, NULL, NULL)))
	{
		free (c);
		return NULL;
	}

	c->fd       = fd;
	c->host     = host;
	c->port     = port;
	c->outgoing = outgoing;
	c->udata    = udata;

	return c;
}

static void tcp_free (TCPConn *c)
{
	if (!c)
		return;

	fdbuf_free (c->buf);
	assert (c->wqueue == NULL);

	free (c);
}

/*****************************************************************************/

TCPConn *tcp_open (in_addr_t host, in_port_t port, int block)
{
	TCPConn *c;
	int      fd;

	if ((fd = net_connect (net_ip_str (host), port, block)) < 0)
		return NULL;

	/* create the structure */
	if (!(c = tcp_new (fd, host, port, TRUE, NULL)))
		net_close (fd);

	return c;
}

TCPConn *tcp_accept (TCPConn *listening, int block)
{
	TCPConn *c;
	int      fd;

	if (!listening)
		return NULL;

	if ((fd = net_accept (listening->fd, block)) < 0)
		return NULL;

	if (!(c = tcp_new (fd, net_peer (fd), listening->port, FALSE, NULL)))
		net_close (fd);

	return c;
}

TCPConn *tcp_bind (in_port_t port, int block)
{
	int fd;

	if (port == 0)
		return NULL;

	if ((fd = net_bind (port, block)) < 0)
		return NULL;

	return tcp_new (fd, 0, port, FALSE, NULL);
}

/*****************************************************************************/

static void finish_queue (TCPConn *c)
{
	input_remove (c->wqueue_id);
	c->wqueue_id = 0;

	unset (&c->wqueue);
}

static int shift_queue (TCPConn *c, int write)
{
	struct write_msg *msg;

	/* grab the next waiting message if available (otherwise we should
	 * assume the queue is now empty instead of a fatal error) */
	if (!(msg = shift (&c->wqueue)))
	{
		finish_queue (c);
		return FALSE;
	}

	assert (msg->data != NULL);
	assert (msg->len > 0);

	if (write)
		net_send (c->fd, (char *)msg->data, msg->len);

	free (msg->data);
	free (msg);

	return TRUE;
}

static void shift_queue_cb (int fd, input_id id, TCPConn *c)
{
	assert (c->fd == fd);
	assert (c->wqueue_id == id);

	shift_queue (c, TRUE);
}

static int push_queue (TCPConn *c, struct write_msg *msg)
{
	if (!push (&c->wqueue, msg))
	{
		TRACE (("eep!"));
		return FALSE;
	}

	if (c->wqueue_id == 0)
	{
		c->wqueue_id = input_add (c->fd, c, INPUT_WRITE,
		                          (InputCallback)shift_queue_cb, FALSE);

		assert (c->wqueue_id > 0);
	}

	return TRUE;
}

/*****************************************************************************/

void tcp_close (TCPConn *c)
{
	if (!c)
		return;

	tcp_flush (c, FALSE);              /* destroy pending writes */
	input_remove_all (c->fd);          /* remove all event inputs */
	net_close (c->fd);                 /* close the socket */
	tcp_free (c);                      /* destroy the data */
}

void tcp_close_null (TCPConn **c)
{
	if (!c || !(*c))
		return;

	tcp_close (*c);
	*c = NULL;
}

/*****************************************************************************/

int tcp_flush (TCPConn *c, int write)
{
	int cnt = 0;

	if (!c)
		return 0;

	while (shift_queue (c, write))
		cnt++;

	/* just in case */
	finish_queue (c);

	return cnt;
}

static int push_msg (TCPConn *c, unsigned char *data, size_t len)
{
	struct write_msg *msg;

	if (!(msg = malloc (sizeof (struct write_msg))))
		return -1;

	/* copy the requested write memory */
	if (!(msg->data = malloc (len)))
	{
		free (msg);
		return -1;
	}

	memcpy (msg->data, data, len);
	msg->len = len;

	push_queue (c, msg);

	return len;
}

int tcp_write (TCPConn *c, unsigned char *data, size_t len)
{
	if (!c || c->fd < 0)
		return -1;

	if (len == 0)
		return 0;

	return push_msg (c, data, len);
}

int tcp_writestr (TCPConn *c, char *data)
{
	return tcp_write (c, (unsigned char *)data, STRLEN (data));
}

int tcp_send (TCPConn *c, unsigned char *data, size_t len)
{
	if (!c || c->fd < 0)
		return -1;

	if (len == 0)
		return 0;

	return net_send (c->fd, (char *)data, len);
}

/*****************************************************************************/

FDBuf *tcp_readbuf (TCPConn *c)
{
	if (!c)
		return NULL;

	return c->buf;
}
