/*
 * $Id: connection.h,v 1.20 2003/03/07 02:34:06 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#ifndef __CONNECTION_H
#define __CONNECTION_H

/*****************************************************************************/

/**
 * @file connection.h
 *
 * @brief High-level TCP connection abstraction.
 *
 * This structure is passed along to pretty much every internal giFT function
 * to encompass all the network functionality that will be required. Also the
 * recommended method of interfacing with the event loop.
 */

/*****************************************************************************/

#include "fdbuf.h"

/*****************************************************************************/

#ifdef __cplusplus
extern "C"
{
#endif /* __reallycrappylanguage */

/*****************************************************************************/

/**
 * Basic building block for all connections spawned using libgiFT.  If you
 * wish to use libgiFT at all, use these.
 */
typedef struct tcp_conn
{
	/**
	 * @name Public
	 */
	int               fd;              /**< The file discriptor of this
	                                    *   connection */
	in_addr_t         host;            /**< Peer host address */
	in_port_t         port;            /**< Peer connection port */
	unsigned char     outgoing : 1;    /**< Simple convenience so you know
										*   which direction this connection
										*   was initiated */
	FDBuf            *buf;             /**< Read buffer */
	void             *udata;           /**< Any arbitrary data that you want
	                                    *   passed along */

	/**
	 * @name Private
	 */
	Array            *wqueue;          /**< Write queue for ::tcp_write */
	input_id          wqueue_id;       /**< Automagically write when avail */
  	float             adjust;          /**< Ratio to adjust the send/recv
	                                    *   buffer */
} TCPConn;

/**
 * Deprecated type definition.  This will be removed as soon as I get around
 * to a major name change.
 */
typedef TCPConn Connection;

/*****************************************************************************/

/**
 * Creates a new connection structure and attempts an outgoing TCP
 * connection using ::net_connect.
 *
 * @see net_connect
 *
 * @return Pointer to a dynamically allocated Connection structure.
 */
TCPConn *tcp_open (in_addr_t host, in_port_t port, int block);

/**
 * Accepts an incoming socket connection.
 *
 * @see tcp_open
 * @see net_accept
 */
TCPConn *tcp_accept (TCPConn *listening, int block);

/**
 * Binds to the supplied port.
 *
 * @see tcp_open
 * @see net_accept
 * @see net_bind
 */
TCPConn *tcp_bind (in_port_t port, int block);

/**
 * Close (and free) a connection pointer.
 *
 * @param c  Location of the Connection structure.
 */
void tcp_close (TCPConn *c);

/**
 * Wrapper for ::tcp_close that sets the calling argument to NULL
 * afterwards.
 *
 * @param c  Address of a Connection pointer.
 */
void tcp_close_null (TCPConn **c);

/*****************************************************************************/

/**
 * Flush and destroy the pending write queue.
 *
 * @param c
 * @param write  If TRUE, appropriate send calls will be performed.  Otherwise,
 *               the data will simply be destroyed.
 *
 * @return Number of elements removed.
 */
int tcp_flush (TCPConn *c, int write);

/**
 * Abstracted send call implemented with a user-space socket write queue to
 * avoid flooding the kernel.
 */
int tcp_write (TCPConn *c, unsigned char *data, size_t len);

/**
 * Simple wrapper for ::tcp_write for NUL-terminated strings.
 */
int tcp_writestr (TCPConn *c, char *data);

/**
 * Direct send wrapper to maintain a consistent interface.
 */
int tcp_send (TCPConn *c, unsigned char *data, size_t len);

/**
 * Accessor for the buffer object.
 */
FDBuf *tcp_readbuf (TCPConn *c);

/*****************************************************************************/

#ifdef __cplusplus
}
#endif /* __reallycrappylanguage */

/*****************************************************************************/

#endif /* __CONNECTION_H */
