/*
 * $Id: event.h,v 1.15 2003/03/07 02:34:06 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#ifndef __EVENT_H
#define __EVENT_H

/*****************************************************************************/

/**
 * @file event.h
 *
 * @brief Main select event loop.
 *
 * Provides the low-level framework for giFT file descriptor (yes, that
 * includes sockets) handling.
 */

/*****************************************************************************/

#include <sys/types.h>

/*****************************************************************************/

#ifdef __cplusplus
extern "C"
{
#endif /* __reallycrappylanguage */

/*****************************************************************************/

/* time specifications useful for timer_add calls, which are in msec
 * precision */
#define MSEC     (1)
#define SECONDS  (1000*MSEC)
#define MINUTES  (60*SECONDS)

/* time specs useful for comparing seconds since the epoch */
#define ESECONDS (1)
#define EMINUTES (60*ESECONDS)
#define EHOURS   (60*EMINUTES)
#define EDAYS    (24*EHOURS)

/*****************************************************************************/

typedef unsigned int input_id;         /**< Timer handle */
typedef unsigned int timer_id;         /**< Input handle */

/*****************************************************************************/

/**
 * Callback used to indicate that the connection has changed state according
 * to the ::input_add state.
 */
typedef void (*InputCallback) (int fd, input_id id, void *udata);

/**
 * Available select sets.
 */
typedef enum
{
	INPUT_READ  = 0x01,
	INPUT_WRITE = 0x02,
	INPUT_ERROR = 0x04
} InputState;

/**
 * File descriptor handler.
 */
typedef struct
{
	int           fd;                  /**< file descriptor */
	input_id      id;
	input_id      poll_id;
	InputState    set;

	InputCallback cb;
	void         *udata;               /**< arbitrary data associated with
	                                    *   this input fd */
	timer_id      validate;            /**< timeout for incomplete sockets */
	unsigned char complete : 1;        /**< complete is FALSE until first
	                                    *   state change.  used for
	                                    *   automated connection timeouts. */
	unsigned char suspended : 1;       /**< suspended is TRUE when input
	                                    *   should not be put in select loop */
	signed char   dirty : 2;           /**< write queue */
} Input;

/*****************************************************************************/

/**
 * Timer has reached zero.
 *
 * @param udata Arbitrary user data.
 *
 * @retval FALSE Clear and remove from event loop.
 * @retval TRUE  Reset this timer.
 */
typedef int (*TimerCallback) (void *udata);

/**
 * Timer event.
 */
struct timer
{
	timer_id       id;                 /**< reference identifier */

	struct timeval expiration;         /**< exact time this timer will
	                                    *   expire */
	struct timeval interval;           /**< interval this timer is set for */

	TimerCallback  cb;
	void          *udata;              /**< arbitrary user data */
};

#ifndef IMAGE_MAGICK_SUCKS
typedef struct timer Timer;
#endif /* IMAGE_MAGICK_SUCKS */

/*****************************************************************************/

/**
 * Initialize the event subsystem.  This is mandatory and must be called
 * prior to any events being scheduled into the main loop.
 */
void event_init (void);

/**
 * Clear all events and return from the main loop.  This will not exit your
 * program, merely break out of ::event_loop, which should be the core
 * blocking entity.
 *
 * @param sig  Signal delivered that caused this interruption.  This value is
 *             currently unused and may disappear entirely.
 */
void event_quit (int sig);

/**
 * Execute a single blocking poll and handle exactly one event.
 */
void event_poll_once (void);

/**
 * Begin the main event loop.  This will block until there are no fds or
 * timers remaining or until ::event_quit has been called.
 */
void event_loop (void);

/*****************************************************************************/

/**
 * Add a new fd input.
 *
 * @param fd
 * @param udata     Arbitrary user data to pass along to \em callback.
 * @param state     Set to examine.
 * @param callback  Function to raise when the examined set detects a change.
 * @param timeout   If TRUE, the fd will be checked for input.  If no response
 *                  comes back in 60 seconds, the fd will be closed and
 *                  the input removed.  \em callback will be called after
 *                  the socket has closed.
 *
 * @return Unique input handle.  You may use this to remove a specific input
 *         or use the file descriptor to remove a group of inputs.
 */
input_id input_add (int fd, void *udata, InputState state,
                   InputCallback callback, int timeout);

/**
 * Remove a single input by it's unique identifier.
 */
void input_remove (input_id id);

/**
 * Remove all inputs registered from a single file descriptor.  This is
 * useful when you want to lazily clean up all your slop, as giFT certainly
 * desires.  The usage of this function is highly discouraged unless you know
 * that the socket is no longer connected as it will interfere with queue
 * systems.
 */
void input_remove_all (int fd);

/**
 * Suspend all inputs registered to this connection from the select
 * loop.  This is useful for bandwidth throttling in user space and some
 * other evil things.
 */
void input_suspend (int fd);

/**
 * Check if the given fd has suspended inputs.
 */
int input_suspended (int fd);

/**
 * Move the inputs back into the select loop.  See ::input_suspend.
 */
void input_resume (int fd);

/*****************************************************************************/

/**
 * Install a new timer event.
 *
 * @param interval  Time to wait.  Expressed in milliseconds.  See ::SECONDS
 *                  and ::MINUTES for helpers.
 * @param callback  Function to be called when the interval is satisfied.
 * @param udata     Arbitrary user data.
 *
 * @return Unique timer identifier on success, otherwise zero.
 */
timer_id timer_add (time_t interval, TimerCallback callback, void *udata);

/**
 * Reset a timer.  That is, reset the expiration time to now + interval as if
 * the timer was registered for the first time.
 */
void timer_reset (timer_id id);

/**
 * Remove a timer.  This clears all data associated with the timer.
 * The callback will not be raised.
 */
void timer_remove (timer_id id);

/**
 * Wrapper for ::timer_remove which assigns the value at id to zero after
 * removal to invalidate the identifier.
 */
void timer_remove_zero (timer_id *id);

/*****************************************************************************/

#ifdef __cplusplus
}
#endif /* __reallycrappylanguage */

/*****************************************************************************/

#endif /* __EVENT_H */
